FIELD_IS_REQUIRED = "Field is required."
ANY_FIELD_IS_REQUIRED = "One of these fields must be provided."
DATE_IS_NOT_AFTER = "Date is too late."
DATE_IS_NOT_BEFORE = "Date is too early."
FIELD_MUST_BE_BLANK = "Field must be blank."



import unittest

import pylidator
from pylidator.exceptions import ContextNotAvailableError

from functools import wraps


def child_generator(obj):
    for row in obj.children:
        yield row


class TestObj(object):
    def __init__(self, returns):
        self.it_happened = False
        self.returns = returns
        self.children = []


@pylidator.validator(of="base_obj")
def validate_parent(obj):
    obj.it_happened = True
    return obj.returns


@pylidator.validator(of="child_obj")
def validate_child(obj_child):
    obj_child.it_happened = True
    return obj_child.returns


@pylidator.validator(of="base_obj", requires="constants_service")
def validate_parent_with_constants_service(obj_child, constants_service):
    obj_child.it_happened = True
    return obj_child.returns


class MyContext:
    pass


def _provide_base_obj(base_obj):
    yield base_obj, None


def _provide_child_obj(base_obj):
    for i, c in enumerate(base_obj.children):
        yield c, {"description": "Child {}".format(i)}


_providers = {"base_obj": _provide_base_obj, "child_obj": _provide_child_obj}


class TestPylidator(unittest.TestCase):
    def test_validator_returns_None_results_in_no_error(self):
        data = TestObj(returns=None)
        ret = pylidator.validate(data, {pylidator.ERROR: [validate_parent]}, providers=_providers)

        self.assertTrue(data.it_happened)
        self.assertEqual([], ret.get_full_results())

    def test_validator_returns_string_results_in_error(self):
        data = TestObj(returns="failed.")
        ret = pylidator.validate(data, {pylidator.ERROR: [validate_parent]}, providers=_providers)
        self.assertTrue(data.it_happened)
        self.assertEqual([{"level": "ERROR", "message": "failed.", "validation_type": None}], ret.get_full_results())

    def test_validator_returns_array_of_strings_results_in_errors(self):
        data = TestObj(returns=["error one", "error two"])
        ret = pylidator.validate(data, {pylidator.ERROR: [validate_parent]}, providers=_providers)
        self.assertEqual(
            [
                {"level": "ERROR", "message": "error one", "validation_type": None},
                {"level": "ERROR", "message": "error two", "validation_type": None},
            ],
            ret.get_full_results(),
        )

    def test_child_validator_returns_string_results_in_error_per_child(self):
        data = TestObj(returns="who cares?")

        data.children.append(TestObj(returns="hi"))
        data.children.append(TestObj(returns=["there", "you"]))
        data.children.append(TestObj(returns={"field1": ["Error 1", "Error 2"], "field2": "Error 3"}))
        data.children.append(TestObj(returns=None))

        ret = pylidator.validate(data, {pylidator.ERROR: [validate_child]}, providers=_providers)
        import pprint

        pprint.pprint(ret.get_full_results())
        self.assertEqual(
            [
                {"description": "Child 0", "level": "ERROR", "message": "hi", "validation_type": None},
                {"description": "Child 1", "level": "ERROR", "message": "there", "validation_type": None},
                {"description": "Child 1", "level": "ERROR", "message": "you", "validation_type": None},
                {
                    "description": "Child 2",
                    "field": "field1",
                    "level": "ERROR",
                    "message": "Field1: ['Error 1', 'Error 2']",
                    "validation_type": None,
                    "verbose_name": "Field1",
                },
                {
                    "description": "Child 2",
                    "field": "field2",
                    "level": "ERROR",
                    "message": "Field2: Error 3",
                    "validation_type": None,
                    "verbose_name": "Field2",
                },
            ],
            ret.get_full_results(),
        )

    def test_validator_with_constants_service_returns_string_results_in_error(self):
        data = TestObj(returns="failed.")
        cs = MyContext()
        ret = pylidator.validate(
            data,
            {pylidator.ERROR: [validate_parent_with_constants_service]},
            providers=_providers,
            extra_context={"constants_service": cs},
        )
        self.assertTrue(data.it_happened)
        self.assertEqual([{"level": "ERROR", "message": "failed.", "validation_type": None}], ret.get_full_results())

    def test_validator_requesting_unavailable_context_throws(self):
        data = TestObj(returns="failed.")
        cs = MyContext()
        with self.assertRaises(ContextNotAvailableError):
            ret = pylidator.validate(
                data,
                {pylidator.ERROR: [validate_parent_with_constants_service]},
                providers=_providers,
                extra_context={"not_constants_service": cs},
            )



from . import messages
from .utils import yield_all

from six import string_types


def any_field_must_be_set(obj, attr, errors):
    """
    Call this method from a `pylidator.validator` to check that at least one of the attributes in `attr` are not None.

    Will add validation messages to `errors` if problems found.

    `obj`  object to test
    `attr` string or iterable of strings to specify attributes to test
    `errors` result list to add any validation messages
    """

    if isinstance(attr, string_types):
        field_must_be_set(obj, attr, errors)
    else:
        for item in attr:
            val = getattr(obj, item)
            if val is not None:
                return

        # All are None, so all get errors.
        errors.append({tuple(attr): messages.ANY_FIELD_IS_REQUIRED})


def field_must_be_set(obj, attr, errors):
    """
    Call this method from a `pylidator.validator` to check that all attributes in `attr` are not None.

    Will add validation messages to `errors` for every None found.

    `obj`  object to test
    `attr` string or iterable of strings to specify attributes to test
    `errors` result list to add any validation messages
    """
    for item in yield_all(attr):
        val = getattr(obj, attr)
        if val is None:
            errors.append({attr: messages.FIELD_IS_REQUIRED})


def field_must_be_none(obj, attr, errors):
    """
    Call this method from a `pylidator.validator` to check that all attributes in `attr` are None.

    Will add validation messages to `errors` for every None found.

    `obj`  object to test
    `attr` string or iterable of strings to specify attributes to test
    `errors` result list to add any validation messages
    """
    for item in yield_all(attr):
        val = getattr(obj, attr)
        if val is not None:
            errors.append({attr: messages.FIELD_MUST_BE_BLANK})


def date_is_not_after(obj, attr, errors, now, allow_none=False, is_inclusive=False):
    val = getattr(obj, attr)
    if val is None:
        if not allow_none:
            errors.append({attr: messages.FIELD_IS_REQUIRED})
        return

    if (val == now and not is_inclusive) or val > now:
        errors.append({attr: messages.DATE_IS_NOT_AFTER})


def date_is_not_before(obj, attr, errors, now, allow_none=False, is_inclusive=False):
    val = getattr(obj, attr)
    if val is None:
        if not allow_none:
            errors.append({attr: messages.FIELD_IS_REQUIRED})
        return

    if (val == now and not is_inclusive) or val < now:
        errors.append({attr: messages.DATE_IS_NOT_BEFORE})



import unittest

import pylidator
from pylidator.exceptions import ContextNotAvailableError

from functools import wraps


def child_generator(obj):
    for row in obj.children:
        yield row


class TestObj(object):
    def __init__(self, returns):
        self.it_happened = False
        self.returns = returns
        self.children = []


@pylidator.validator(of="base_obj")
def validate_parent(obj):
    obj.it_happened = True
    return obj.returns


@pylidator.validator(of="child_obj")
def validate_child(obj_child):
    obj_child.it_happened = True
    return obj_child.returns


@pylidator.validator(of="base_obj", requires="constants_service")
def validate_parent_with_constants_service(obj_child, constants_service):
    obj_child.it_happened = True
    return obj_child.returns


class MyContext:
    pass


def _provide_base_obj(base_obj):
    yield base_obj, None


def _provide_child_obj(base_obj):
    for i, c in enumerate(base_obj.children):
        yield c, {"description": "Child {}".format(i)}


_providers = {"base_obj": _provide_base_obj, "child_obj": _provide_child_obj}


class Tests(unittest.TestCase):
    def test_field_name_mapper_with_children_changes_field_name(self):
        def _custom_field_name_mapper(obj, field_name):
            return field_name + "_mapped"

        data = TestObj(returns={"field1": ["Error 1", "Error 2"], "field2": "Error 3"})

        data.children.append(TestObj(returns={"field1": "Broken"}))
        data.children.append(TestObj(returns={"field1": ["Error 1", "Error 2"], "field2": "Error 3"}))
        data.children.append(TestObj(returns=None))

        ret = pylidator.validate(
            data,
            {pylidator.ERROR: [validate_parent, validate_child]},
            providers=_providers,
            field_name_mapper=_custom_field_name_mapper,
        )

        self.assertEqual(
            [
                {
                    "field": "field1",
                    "level": "ERROR",
                    "message": "field1_mapped: ['Error 1', 'Error 2']",
                    "validation_type": None,
                    "verbose_name": "field1_mapped",
                },
                {
                    "field": "field2",
                    "level": "ERROR",
                    "message": "field2_mapped: Error 3",
                    "validation_type": None,
                    "verbose_name": "field2_mapped",
                },
                {
                    "description": "Child 0",
                    "field": "field1",
                    "level": "ERROR",
                    "message": "field1_mapped: Broken",
                    "validation_type": None,
                    "verbose_name": "field1_mapped",
                },
                {
                    "description": "Child 1",
                    "field": "field1",
                    "level": "ERROR",
                    "message": "field1_mapped: ['Error 1', 'Error 2']",
                    "validation_type": None,
                    "verbose_name": "field1_mapped",
                },
                {
                    "description": "Child 1",
                    "field": "field2",
                    "level": "ERROR",
                    "message": "field2_mapped: Error 3",
                    "validation_type": None,
                    "verbose_name": "field2_mapped",
                },
            ],
            ret.get_full_results(),
        )




import logging

from six import string_types
from .constants import ERROR, WARN, LEVELS

logger = logging.getLogger(__name__)


class Error(dict):
    ERROR = ERROR
    WARN = WARN

    LEVELS = [ERROR, WARN]


class ErrorLedger(object):
    ERROR = ERROR
    WARN = WARN

    def __init__(self, default_object_data=None, logging=True):
        """
        `custom_field_name_mapper` is an optional callable that takes the field name and returns
            a verbose_name for the field.
        """
        self._errors = []
        self._is_valid = True
        self._default_object_data = default_object_data if default_object_data is not None else {}
        self._logging = logging
        self._already_logged = set()

    @staticmethod
    def create_error_object(message, level, object_data=None):
        if isinstance(message, dict):
            assert len(message) == 1, "Don't currently support multi key dicts inside lists."

            for field_name, error in message.items():
                error = "{}: {}".format(field_name, error)
                new_item = Error({"message": error, "field": field_name})
                break

        elif isinstance(message, string_types):
            new_item = Error({"message": message})

        else:
            raise ValueError("Message is required and must be a string or a dict: {}".format(message))

        new_item["level"] = level
        if object_data:
            new_item.update(object_data)

        return new_item

    def merge_with(self, other_ledger):
        assert isinstance(other_ledger, ErrorLedger)

        self._errors += other_ledger._errors
        self._is_valid = self._is_valid and other_ledger._is_valid
        self._already_logged = self._already_logged.union(other_ledger._already_logged)

    def add_message(self, message, level, object_data=None):
        new_item = self.create_error_object(message, level, object_data)
        new_item.update(self._default_object_data)
        self.add_object(new_item)

    def add_object(self, new_item_data):
        new_item = Error()
        new_item.update(self._default_object_data)
        new_item.update(new_item_data)

        new_item["level"]
        new_item["message"]

        if self._logging:
            log_message = "{} {}".format(new_item["level"], new_item["message"])
            # It is annoying when it writes the same msg a million times...
            if log_message not in self._already_logged:
                logger.debug(log_message)
                self._already_logged.add(log_message)

        self._errors.append(new_item)
        if new_item["level"] == self.ERROR:
            self._is_valid = False

    def get_full_results(self):
        return self._errors

    def get_error_messages(self):
        return list(unique_everseen([e["message"] for e in self._errors if e["level"] == e.ERROR]))

    def get_descriptive_error_messages(self):
        return list(["{} {}".format(e.get("description"), e["message"]) for e in self._errors if e["level"] == e.ERROR])

    def get_warning_messages(self):
        return list(unique_everseen([e["message"] for e in self._errors if e["level"] == e.WARN]))

    def get_errors(self):
        return [e for e in self._errors if e["level"] == e.ERROR]

    def get_warnings(self):
        return [e for e in self._errors if e["level"] == e.WARN]

    def is_valid(self):
        return self._is_valid

    def get_django_validation_formatted_errors(self):
        ret = {}
        for e in self._errors:
            field = e.get("field", "non_field_errors")
            level = e["level"]

            try:
                d = ret[level]
            except KeyError:
                d = ret[level] = {}

            try:
                d2 = d[field]
            except KeyError:
                d2 = d[field] = []

            prefixToRemove = e.get("verbose_name", "") + ": "
            if e["message"].startswith(prefixToRemove):
                d2.append(e["message"][len(prefixToRemove) :])
            else:
                d2.append(e["message"])

        return ret


def unique_everseen(iterable, key=None):
    "List unique elements, preserving order. Remember all elements ever seen."
    # unique_everseen('AAAABBBCCDAABBB') --> A B C D
    # unique_everseen('ABBCcAD', str.lower) --> A B C D
    from itertools import filterfalse

    seen = set()
    seen_add = seen.add
    if key is None:
        for element in filterfalse(seen.__contains__, iterable):
            seen_add(element)
            yield element
    else:
        for element in iterable:
            k = key(element)
            if k not in seen:
                seen_add(k)
                yield element



import logging
from functools import wraps
from six import string_types

from .error_ledger import Error, ErrorLedger
from . import exceptions

logger = logging.getLogger(__name__)


def validate(
    obj,
    validators=None,
    providers=None,
    extra_context=None,
    field_name_mapper=None,
    validation_type=None,
    logging=True,
    why="",
):
    """
    `obj` is the top-level object requiring validation.
    `validators` is a dict of {level: list of `@pylidator.validator` objects}
    `providers` is a dict of {of: func that takes obj and returns an iterable of some subobjects}
    `extra_context` is a dict of other data that can be injected into `@pylidator.validator` with `requires`.
    `field_name_mapper` is a string->string func that converts field names given in returned errors into verbose names.
    `validation_type` is added into the error object.
    `logging` If set to False, disables logging of validation results.
    `why` String added to logging to identify the logpoint.
    """

    ledger = ErrorLedger(default_object_data={"validation_type": validation_type}, logging=logging)

    def _process_validator_results(ret, level, object_data, obj):
        """ Process the return of a user-supplied `validator`.  Accepts lists, dicts, and strings. """

        # The first object in the tuple is the one being validated
        if isinstance(obj, tuple):
            real_obj = obj[0]
        else:
            real_obj = obj

        if not ret:
            is_valid = True
            return is_valid

        if isinstance(ret, string_types):
            ledger.add_message(ret, level, object_data)
            is_valid = False

        elif isinstance(ret, dict):
            for field_name, error in list(ret.items()):
                # verbose_field_name = ledger.map_field_name_to_verbose_name(obj, field_name)
                object_data_with_field = object_data.copy()
                object_data_with_field["field"] = field_name
                if field_name_mapper is None:
                    # raise RuntimeError("A field_name_mapper was not supplied to this validator.")
                    verbose_name = None
                else:
                    verbose_name = field_name_mapper(real_obj, field_name)
                if verbose_name is None:
                    from titlecase import titlecase

                    verbose_name = titlecase(" ".join(field_name.split("_")))

                object_data_with_field["verbose_name"] = verbose_name
                error = "{}: {}".format(verbose_name, error)
                ledger.add_message(error, level, object_data_with_field)
                is_valid = False

        else:
            for validator_ret_item in ret:
                if isinstance(validator_ret_item, str):
                    ledger.add_message(validator_ret_item, level, object_data)
                    is_valid = False
                elif isinstance(validator_ret_item, dict):
                    for field_name, error in list(validator_ret_item.items()):
                        # verbose_field_name = ledger.map_field_name_to_verbose_name(obj, field_name)
                        object_data_with_field = object_data.copy()
                        object_data_with_field["field"] = field_name
                        verbose_name = field_name_mapper(real_obj, field_name)
                        if verbose_name is None:
                            from titlecase import titlecase

                            verbose_name = titlecase(" ".join(field_name.split("_")))

                        object_data_with_field["verbose_name"] = verbose_name
                        error = "{}: {}".format(verbose_name, error)
                        ledger.add_message(error, level, object_data_with_field)
                        is_valid = False

        return is_valid

    global _cached_provided_items
    _cached_provided_items = {None: [obj]}

    def get_provided_items(of):
        global _cached_provided_items

        if of in _cached_provided_items:
            return _cached_provided_items[of]

        # Already in cache...
        # if of is None:
        #     return [obj]

        # Use the correct `provider` to find the child object to process with the validator func.
        # The provider will generate all child objects and call the validator func once per yielded object.
        try:
            generator = providers[of]
        except (KeyError, TypeError):
            raise KeyError("Must add `{}` to providers for validator `{}`.".format(of, validator_func_name))

        ret = tuple(generator(obj))
        _cached_provided_items[of] = ret
        return ret

    validator_func_kwargs = {
        "process_validator_results": _process_validator_results,
        "extra_context": extra_context,
        "get_provided_items_f": get_provided_items,
    }

    validators_applied = []
    for level, level_validators in validators.items():
        assert level in Error.LEVELS, "Level `{}` is not recognized.".format(level)

        for v in unique_everseen(level_validators):
            is_valid = v(level=level, **validator_func_kwargs)
            validators_applied.append("{} {}".format(v.__name__, "OK" if is_valid else str(level)))

    if logging:
        if why:
            why = why.strip() + ": "
        else:
            why = ""
        if len(ledger.get_errors()) + len(ledger.get_warnings()) == 0:
            logger.debug(
                "{}validate complete ({} err, {} warn)".format(
                    why, len(ledger.get_errors()), len(ledger.get_warnings())
                )
            )
        else:
            logger.debug(
                "{}validate complete ({} err, {} warn): {}".format(
                    why, len(ledger.get_errors()), len(ledger.get_warnings()), ", ".join(validators_applied)
                )
            )

    return ledger


def validator(of, requires=None, affects=None):
    """
    Decorator for marking validator functions.

    `of` specifies what provider the validator should use.   The `validate` call needs an item in `providers`
         that matches `of`.
    `requires` (optional) can add additional context items, such as the current time or other services that can supply
         data or settings to the validator.  The requirement is fulfilled by passing `extra_context` to the `validate`
         call, containing any items that are used in a `requires`.
    `affects` (optional) is simply passed through to results.  It can be used as guidance for UI/error reporting for
         helping to resolve any resultant errors.
    """

    def decorator(validator_func):
        validator_func_name = validator_func.__name__

        @wraps(validator_func)
        def actually_run_validator_func(process_validator_results, get_provided_items_f, extra_context, level):
            """ `actually_run_validator_func` gets called directly from `validate` above, once per unique validation method in
            `validators`.
            """
            kwargs = {}
            if requires:
                if isinstance(requires, string_types):
                    requires_list = requires.split()
                else:
                    requires_list = requires
                for extra_context_item in requires_list:
                    try:
                        kwargs[extra_context_item] = extra_context[extra_context_item]
                    except (KeyError, TypeError) as exc:
                        raise exceptions.ContextNotAvailableError(
                            "{} is not available in the validator context.".format(extra_context_item)
                        )

            # logger.debug(u'Validating {} of {} (of={}).'.format(validator_func, obj, of))
            is_valid = True
            for item in get_provided_items_f(of):
                try:
                    row, object_data = item
                except ValueError:
                    raise ValueError("{} must yield 2-tuples, got {}".format(generator, item))
                assert object_data is None or isinstance(
                    object_data, dict
                ), "Object data returned from provider must be None or dict of values, but got {}".format(
                    type(object_data)
                )

                if object_data is None:
                    object_data = {}

                if affects:
                    object_data["affects"] = affects

                ret = validator_func(row, **kwargs)
                row_is_valid = process_validator_results(ret, level=level, object_data=object_data, obj=row)
                if not row_is_valid:
                    is_valid = False

            return is_valid

        return actually_run_validator_func

    return decorator


def format_results(validator_results):
    if validator_results.is_valid():
        return "is valid."

    def format_result_item(err):
        err = err.copy()

        level = err.pop("level")
        message = err.pop("message")
        description = err.pop("description", "(no description)")

        therest = ", ".join("{}={}".format(x, y) for x, y in err.items())
        return "{} {} {} {}".format(level, description, message, therest)

    return "\n".join((format_result_item(err) for err in validator_results.get_full_results()))


def unique_everseen(iterable, key=None):
    "List unique elements, preserving order. Remember all elements ever seen."
    # unique_everseen('AAAABBBCCDAABBB') --> A B C D
    # unique_everseen('ABBCcAD', str.lower) --> A B C D
    from future.moves.itertools import filterfalse

    seen = set()
    seen_add = seen.add
    if key is None:
        for element in filterfalse(seen.__contains__, iterable):
            seen_add(element)
            yield element
    else:
        for element in iterable:
            k = key(element)
            if k not in seen:
                seen_add(k)
                yield element



from six import string_types


def yield_all(str_or_iter):
    # http://stackoverflow.com/a/11106461/237091
    def list_like(v):
        """Return True if v is a non-string sequence and is iterable. Note that
       not all objects with getitem() have the iterable attribute"""
        if hasattr(v, "__iter__") and not isinstance(v, string_types):
            return True
        else:
            # This will happen for most atomic types like numbers and strings
            return False

    if list_like(str_or_iter):
        for item in str_or_iter:
            yield item
    else:
        yield str_or_iter
